// =====================================================================
//  filter.c
// =====================================================================

//  Author:         (c) 2013 Camil Demetrescu
//  License:        See the end of this file for license information
//  Created:        July 7, 2013

//  Last changed:   $Date: 2013/07/07 15:00:07 $
//  Changed by:     $Author: demetres $
//  Revision:       $Revision: 1.00 $


#include "filter.h"
#include <stdint.h>
#include <string.h>

// memcpy
#ifndef USEMEMCPY
#define USEMEMCPY 0
#else
  #if USEMEMCPY
  #warning "using memcpy in this implementation"
  #endif
#endif

// platform id
#if UINTPTR_MAX == 0xFFFFFFFF
#warning "32 bit platform detected"
#define __32BIT__
#elif UINTPTR_MAX == 0xFFFFFFFFFFFFFFFF
#warning "64 bit platform detected"
#define __64BIT__
#else
#error "cannot determine platform"
#endif

// macros
#define NEIGHBOR(m,i,dy,dx,w) (*((m)+(i)+(dy)*(w)+(dx)))
#define CONVOLUTION(i) do {                                \
    val  = NEIGHBOR(img_in, (i), -2, -2, cols)*filter[0];  \
    val += NEIGHBOR(img_in, (i), -2, -1, cols)*filter[1];  \
    val += NEIGHBOR(img_in, (i), -2,  0, cols)*filter[2];  \
    val += NEIGHBOR(img_in, (i), -2, +1, cols)*filter[3];  \
    val += NEIGHBOR(img_in, (i), -2, +2, cols)*filter[4];  \
    val += NEIGHBOR(img_in, (i), -1, -2, cols)*filter[5];  \
    val += NEIGHBOR(img_in, (i), -1, -1, cols)*filter[6];  \
    val += NEIGHBOR(img_in, (i), -1,  0, cols)*filter[7];  \
    val += NEIGHBOR(img_in, (i), -1, +1, cols)*filter[8];  \
    val += NEIGHBOR(img_in, (i), -1, +2, cols)*filter[9];  \
    val += NEIGHBOR(img_in, (i),  0, -2, cols)*filter[10]; \
    val += NEIGHBOR(img_in, (i),  0, -1, cols)*filter[11]; \
    val += NEIGHBOR(img_in, (i),  0,  0, cols)*filter[12]; \
    val += NEIGHBOR(img_in, (i),  0, +1, cols)*filter[13]; \
    val += NEIGHBOR(img_in, (i),  0, +2, cols)*filter[14]; \
    val += NEIGHBOR(img_in, (i), +1, -2, cols)*filter[15]; \
    val += NEIGHBOR(img_in, (i), +1, -1, cols)*filter[16]; \
    val += NEIGHBOR(img_in, (i), +1,  0, cols)*filter[17]; \
    val += NEIGHBOR(img_in, (i), +1, +1, cols)*filter[18]; \
    val += NEIGHBOR(img_in, (i), +1, +2, cols)*filter[19]; \
    val += NEIGHBOR(img_in, (i), +2, -2, cols)*filter[20]; \
    val += NEIGHBOR(img_in, (i), +2, -1, cols)*filter[21]; \
    val += NEIGHBOR(img_in, (i), +2,  0, cols)*filter[22]; \
    val += NEIGHBOR(img_in, (i), +2, +1, cols)*filter[23]; \
    val += NEIGHBOR(img_in, (i), +2, +2, cols)*filter[24]; \
  val = val*factor+bias;                                 \
  img_out[i] = (unsigned char)(val < 0 ? 0 :             \
                       val > 255 ? 255 : val);   \
} while(0)

#ifdef DOUBLE_FILTER
typedef double VT;
#else
typedef int VT;
#endif

//#undef __64BIT__
//#define __32BIT__

// param setup
#if defined(__32BIT__)
typedef unsigned int T;
#define RUN   0xFFFFFFFF
#else
typedef unsigned long T;
#define RUN   0xFFFFFFFFFFFFFFFF
#endif

// instrumented version
#if VERSION == 3
#include "kstream.h"
extern kstream* ks;
#endif


// ---------------------------------------------------------------------
// trim_mask
// ---------------------------------------------------------------------
void trim_mask(unsigned char* mask, int rows, int cols) {
    long i;
    memset(mask, 255, 2*cols);
    for (i=2; i<rows-2; i++)
        mask[i*cols] = 
        mask[i*cols+1] = 
        mask[(i+1)*cols-2] = 
        mask[(i+1)*cols-1] = 255;
    memset(mask+(rows-2)*cols, 255, 2*cols);
}


// ---------------------------------------------------------------------
// filter_conv
// ---------------------------------------------------------------------
#if VERSION == 0
#warning "compiling original version"
void filter_conv(unsigned char* img_in, 
                 unsigned char* img_out, 
                 unsigned char* mask, 
                 FT filter[25],
                 double factor, double bias,
                 int rows, int cols) {
    
    VT val;
    long n = rows*cols, i;
    
    #if USEMEMCPY
    memcpy(img_out, img_in, n);
    #endif
    
    for (i = 0; i < n; i++)
        #if USEMEMCPY
        if (!mask[i]) CONVOLUTION(i);
        #else
        if (mask[i]) img_out[i] = img_in[i];
        else CONVOLUTION(i);
        #endif
}

#elif VERSION == 1
#warning "compiling optimized version 1"

#if USEMEMCPY

void filter_conv(unsigned char* img_in, 
                 unsigned char* img_out, 
                 unsigned char* mask, 
                 FT filter[25],
                 double factor, double bias,
                 int rows, int cols) {

    long n = rows*cols, i, j;
    VT val;

    memcpy(img_out, img_in, n);

    for (i = 0; i < n-sizeof(T)+1; i += sizeof(T))
        if (*(T*)(mask+i) == 0) {  
            CONVOLUTION(i);
            CONVOLUTION(i+1);
            CONVOLUTION(i+2);
            CONVOLUTION(i+3);
            #ifdef __64BIT__
            CONVOLUTION(i+4);
            CONVOLUTION(i+5);
            CONVOLUTION(i+6);
            CONVOLUTION(i+7);
            #endif
        }
        else if (*(T*)(mask+i) != RUN)
            for (j = i; j < i+sizeof(T); j++)
                if (!mask[j]) CONVOLUTION(j);
}

#else

void filter_conv(unsigned char* img_in, 
                 unsigned char* img_out, 
                 unsigned char* mask, 
                 FT filter[25],
                 double factor, double bias,
                 int rows, int cols) {
    
    long n = rows*cols, i, j;
    VT val;
    
    for (i = 0; i < n-sizeof(T)+1; i += sizeof(T)) {
        if (*(T*)(mask+i) == RUN)
            *(T*)(img_out+i) = *(T*)(img_in+i);
        else if (*(T*)(mask+i) == 0) {  
            CONVOLUTION(i);
            CONVOLUTION(i+1);
            CONVOLUTION(i+2);
            CONVOLUTION(i+3);
            #ifdef __64BIT__
            CONVOLUTION(i+4);
            CONVOLUTION(i+5);
            CONVOLUTION(i+6);
            CONVOLUTION(i+7);
            #endif
        }
        else for (j = i; j < i+sizeof(T); j++)
            if (!mask[j]) CONVOLUTION(j);
    }
}
#endif

#elif VERSION == 3
#warning "compiling instrumented version"

void filter_conv(unsigned char* img_in, 
                 unsigned char* img_out, 
                 unsigned char* mask, 
                 FT filter[25],
                 double factor, double bias,
                 int rows, int cols) {

    VT val;
    long n = rows*cols, i;

    #if USEMEMCPY
    memcpy(img_out, img_in, n);
    #endif

    kstream_start(ks);
    kstream_emit(ks, 0);
 
    for (i = 0; i < n; i++)        
        if (!mask[i]) {        
            CONVOLUTION(i);
            kstream_emit(ks, 2);                    
        } else {
            #if USEMEMCPY==0
            img_out[i] = img_in[i];
            #endif
            kstream_emit(ks, 1);
        }
}
#endif


// Copyright (C) 2013 Camil Demetrescu

// This library is free software; you can redistribute it and/or
// modify it under the terms of the GNU Lesser General Public
// License as published by the Free Software Foundation; either
// version 2.1 of the License, or (at your option) any later version.

// This library is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
// Lesser General Public License for more details.

// You should have received a copy of the GNU Lesser General Public
// License along with this library; if not, write to the Free Software
// Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307
// USA
